#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "playfair.h"

#define ALPHA "ABCDEFGHIJKLMNOPQRSTUVXYZ"

char *create_matrix(const char *key)
{
    char *matrix = (char *)malloc(25 * sizeof(char));
    if (matrix == NULL)
        return NULL;

    char key_chars[26] = {0};
    int key_len = 0;
    for (int i = 0; key[i] != '\0'; i++)
    {
        char c = toupper(key[i]);
        if (c >= 'A' && c <= 'Z')
        {
            if (key_chars[c - 'A'] == 0)
            {
                key_chars[c - 'A'] = 1;
                key_len++;
            }
        }
    }

    int idx = 0;
    for (int i = 0; i < 26; i++)
    {
        if (key_chars[i] == 1)
        {
            matrix[idx++] = 'A' + i;
        }
    }
    for (int i = 0; i < 26; i++)
    {
        if (key_chars[i] == 0)
        {
            matrix[idx++] = 'A' + i;
        }
    }

    return matrix;
}

void find_position(char *matrix, char c, int *row, int *col)
{
    for (int i = 0; i < 5; i++)
    {
        for (int j = 0; j < 5; j++)
        {
            if (matrix[i * 5 + j] == c)
            {
                *row = i;
                *col = j;
                return;
            }
        }
    }
    *row = -1;
    *col = -1;
}

char *playfair_encrypt(const char *key, const char *text)
{
    if (key == NULL || text == NULL)
    {
        return NULL;
    }

    for (size_t i = 0; key[i] != '\0'; i++)
    {
        if (!isalpha(key[i]))
        {
            return NULL;
        }
    }

    char *matrix = create_matrix(key);
    if (matrix == NULL)
    {
        return NULL;
    }

    char *processed_text = (char *)malloc(strlen(text) + 1);
    if (processed_text == NULL)
    {
        free(matrix);
        return NULL;
    }

    int processed_len = 0;
    for (int i = 0; text[i] != '\0'; i++)
    {
        if (text[i] == ' ')
        {
            continue;
        }
        char c = toupper(text[i]);
        if (c >= 'A' && c <= 'Z')
        {
            processed_text[processed_len++] = c;
        }
    }
    processed_text[processed_len] = '\0';

    char *ciphertext = (char *)malloc((processed_len + 1) * sizeof(char));
    if (ciphertext == NULL)
    {
        free(matrix);
        free(processed_text);
        return NULL;
    }

    int i = 0;
    while (i < processed_len)
    {
        char a = processed_text[i];
        char b = (i + 1 < processed_len) ? processed_text[i + 1] : 'X';

        if (a == b)
        {
            b = 'X';
        }

        int r1, c1, r2, c2;
        find_position(matrix, a, &r1, &c1);
        find_position(matrix, b, &r2, &c2);

        if (r1 == r2)
        {
            ciphertext[i] = matrix[r1 * 5 + (c1 + 1) % 5];
            ciphertext[i + 1] = matrix[r2 * 5 + (c2 + 1) % 5];
        }
        else if (c1 == c2)
        {
            ciphertext[i] = matrix[((r1 + 1) % 5) * 5 + c1];
            ciphertext[i + 1] = matrix[((r2 + 1) % 5) * 5 + c2];
        }
        else
        {
            ciphertext[i] = matrix[r1 * 5 + c2];
            ciphertext[i + 1] = matrix[r2 * 5 + c1];
        }

        i += 2;
    }

    free(matrix);
    free(processed_text);
    return ciphertext;
}

char *playfair_decrypt(const char *key, const char *text)
{
    if (key == NULL || text == NULL)
    {
        return NULL;
    }

    char *matrix = create_matrix(key);
    if (matrix == NULL)
    {
        return NULL;
    }

    char *plaintext = (char *)malloc(strlen(text) + 1);
    if (plaintext == NULL)
    {
        free(matrix);
        return NULL;
    }

    int i = 0;
    while (i < strlen(text))
    {
        char a = text[i];
        char b = text[i + 1];

        int r1, c1, r2, c2;
        find_position(matrix, a, &r1, &c1);
        find_position(matrix, b, &r2, &c2);

        if (r1 == r2)
        {
            plaintext[i] = matrix[r1 * 5 + (c1 - 1 + 5) % 5];
            plaintext[i + 1] = matrix[r2 * 5 + (c2 - 1 + 5) % 5];
        }
        else if (c1 == c2)
        {
            plaintext[i] = matrix[((r1 - 1 + 5) % 5) * 5 + c1];
            plaintext[i + 1] = matrix[((r2 - 1 + 5) % 5) * 5 + c2];
        }
        else
        {
            plaintext[i] = matrix[r1 * 5 + c2];
            plaintext[i + 1] = matrix[r2 * 5 + c1];
        }

        i += 2;
    }

    free(matrix);
    return plaintext;
}